@@ -59,6 +59,10 @@ from dpgen2.flow.dpgen_loop import ConcurrentLearning
 from dpgen2.exploration.report import ExplorationReport
 from dpgen2.exploration.task import ExplorationTaskGroup, ExplorationStage
 from dpgen2.exploration.selector import TrustLevel
+from dpgen2.utils import(
+    dump_object_to_file,
+    load_object_from_file,
+)
 
 from dpgen2.constants import (
     train_task_pattern,
@@ -102,6 +106,14 @@ from mocked_ops import (
     MockedStage2,
     MockedConstTrustLevelStageScheduler,
 )
+from dpgen2.utils.step_config import normalize as normalize_step_dict
+default_config = normalize_step_dict(
+    {
+        "template_config" : {
+            "image" : default_image,
+        }
+    }
+)
 
 
 @unittest.skipIf(skip_ut_with_dflow, skip_ut_with_dflow_reason)
@@ -112,24 +124,24 @@ class TestLoop(unittest.TestCase):
             MockedPrepDPTrain,
             MockedRunDPTrain,
             upload_python_package = upload_python_package,
-            prep_image = default_image,
-            run_image = default_image,
+            prep_config = default_config,
+            run_config = default_config,
         )
         self.prep_run_lmp_op = PrepRunLmp(
             "prep-run-lmp",
             PrepLmp,
             MockedRunLmp,
             upload_python_package = upload_python_package,
-            prep_image = default_image,
-            run_image = default_image,
+            prep_config = default_config,
+            run_config = default_config,
         )
         self.prep_run_fp_op = PrepRunFp(
             "prep-run-fp",
             MockedPrepVasp,
             MockedRunVasp,
             upload_python_package = upload_python_package,
-            prep_image = default_image,
-            run_image = default_image,
+            prep_config = default_config,
+            run_config = default_config,
         )
         self.block_cl_op = ConcurrentLearningBlock(
             self.name+'-block', 
@@ -139,14 +151,14 @@ class TestLoop(unittest.TestCase):
             self.prep_run_fp_op,
             MockedCollectData,
             upload_python_package = upload_python_package,
-            select_confs_image = default_image,
-            collect_data_image = default_image,
+            select_confs_config = default_config,
+            collect_data_config = default_config,
         )        
         self.dpgen_op = ConcurrentLearning(
             self.name,
             self.block_cl_op,
             upload_python_package = upload_python_package,
-            image = default_image,
+            step_config = default_config,
         )
 
     def _setUp_data(self):
@@ -173,7 +185,7 @@ class TestLoop(unittest.TestCase):
         
         self.template_script = mocked_template_script
 
-        self.type_map = []
+        self.type_map = ['H', 'O']
 
         self.incar = Path('incar')
         self.incar.write_text(mocked_incar_template)
@@ -184,6 +196,9 @@ class TestLoop(unittest.TestCase):
             self.incar,
             {'foo': 'potcar'},
         )
+        self.vasp_inputs_fname = Path('vasp_inputs.dat')
+        self.vasp_inputs_arti = upload_artifact(
+            dump_object_to_file(self.vasp_inputs, self.vasp_inputs_fname))
 
         self.scheduler = ExplorationScheduler()        
         self.trust_level = TrustLevel(0.1, 0.3)
@@ -203,9 +218,9 @@ class TestLoop(unittest.TestCase):
             max_numb_iter = 2,
         )
         self.scheduler.add_stage_scheduler(stage_scheduler)        
-        with open('in_scheduler.dat', 'wb') as fp:
-            pickle.dump(self.scheduler, fp)
-        self.scheduler_artifact = upload_artifact('in_scheduler.dat')
+        self.scheduler_artifact = upload_artifact(
+            dump_object_to_file(self.scheduler, 'in_scheduler.dat')
+        )
         
     def setUp(self):
         self.name = 'dpgen'
@@ -221,7 +236,7 @@ class TestLoop(unittest.TestCase):
             name = Path(model_name_pattern % ii)
             if name.is_file():
                 os.remove(name)
-        for ii in [self.incar, self.potcar, Path('scheduler.dat'), Path('in_scheduler.dat')]:
+        for ii in [self.incar, self.potcar, Path('scheduler.dat'), Path('in_scheduler.dat'), self.vasp_inputs_fname]:
             if ii.is_file():
                 os.remove(ii)
 
@@ -251,11 +266,11 @@ class TestLoop(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -273,9 +288,9 @@ class TestLoop(unittest.TestCase):
         self.assertEqual(step.phase, "Succeeded")        
         
         # scheduler = jsonpickle.decode(step.outputs.parameters['exploration_scheduler'].value)
-        download_artifact(step.outputs.artifacts["exploration_scheduler"])
-        with open('scheduler.dat', 'rb') as fp:
-            scheduler = pickle.load(fp)
+        scheduler = load_object_from_file(
+            download_artifact(step.outputs.artifacts["exploration_scheduler"])[0]
+        )
         download_artifact(step.outputs.artifacts["iter_data"], path = 'iter_data')
         download_artifact(step.outputs.artifacts["models"], path = Path('models')/self.name)
         self.assertEqual(scheduler.get_stage(), 2)
@@ -324,40 +339,40 @@ class TestLoopRestart(unittest.TestCase):
             MockedPrepDPTrain,
             MockedRunDPTrain,
             upload_python_package = upload_python_package,
-            prep_image = default_image,
-            run_image = default_image,
+            prep_config = default_config,
+            run_config = default_config,
         )
         self.prep_run_lmp_op = PrepRunLmp(
             "prep-run-lmp",
             PrepLmp,
             MockedRunLmp,
             upload_python_package = upload_python_package,
-            prep_image = default_image,
-            run_image = default_image,
+            prep_config = default_config,
+            run_config = default_config,
         )
         self.prep_run_fp_op = PrepRunFp(
             "prep-run-fp",
             MockedPrepVasp,
             MockedRunVasp,
             upload_python_package = upload_python_package,
-            prep_image = default_image,
-            run_image = default_image,
+            prep_config = default_config,
+            run_config = default_config,
         )
         self.prep_run_fp_op_f1 = PrepRunFp(
             "prep-run-fp",
             MockedPrepVasp,
             MockedRunVaspFail1,
             upload_python_package = upload_python_package,
-            prep_image = default_image,
-            run_image = default_image,
+            prep_config = default_config,
+            run_config = default_config,
         )
         self.prep_run_fp_op_res = PrepRunFp(
             "prep-run-fp",
             MockedPrepVasp,
             MockedRunVaspRestart,
             upload_python_package = upload_python_package,
-            prep_image = default_image,
-            run_image = default_image,
+            prep_config = default_config,
+            run_config = default_config,
         )
         # failed collect data
         self.block_cl_op_0 = ConcurrentLearningBlock(
@@ -368,14 +383,14 @@ class TestLoopRestart(unittest.TestCase):
             self.prep_run_fp_op,
             MockedCollectDataFailed,
             upload_python_package = upload_python_package,
-            select_confs_image = default_image,
-            collect_data_image = default_image,
+            select_confs_config = default_config,
+            collect_data_config = default_config,
         )        
         self.dpgen_op_0 = ConcurrentLearning(
             self.name,
             self.block_cl_op_0,
             upload_python_package = upload_python_package,
-            image = default_image,
+            step_config = default_config,
         )
         # restart collect data
         self.block_cl_op_1 = ConcurrentLearningBlock(
@@ -386,14 +401,14 @@ class TestLoopRestart(unittest.TestCase):
             self.prep_run_fp_op,
             MockedCollectDataRestart,
             upload_python_package = upload_python_package,
-            select_confs_image = default_image,
-            collect_data_image = default_image,
+            select_confs_config = default_config,
+            collect_data_config = default_config,
         )        
         self.dpgen_op_1 = ConcurrentLearning(
             self.name,
             self.block_cl_op_1,
             upload_python_package = upload_python_package,
-            image = default_image,
+            step_config = default_config,
         )
         # failed vasp 1
         self.block_cl_op_2 = ConcurrentLearningBlock(
@@ -404,14 +419,14 @@ class TestLoopRestart(unittest.TestCase):
             self.prep_run_fp_op_f1,
             MockedCollectData,
             upload_python_package = upload_python_package,
-            select_confs_image = default_image,
-            collect_data_image = default_image,
+            select_confs_config = default_config,
+            collect_data_config = default_config,
         )        
         self.dpgen_op_2 = ConcurrentLearning(
             self.name,
             self.block_cl_op_2,
             upload_python_package = upload_python_package,
-            image = default_image,
+            step_config = default_config,
         )
         # restart vasp
         self.block_cl_op_3 = ConcurrentLearningBlock(
@@ -422,14 +437,14 @@ class TestLoopRestart(unittest.TestCase):
             self.prep_run_fp_op_res,
             MockedCollectData,
             upload_python_package = upload_python_package,
-            select_confs_image = default_image,
-            collect_data_image = default_image,
+            select_confs_config = default_config,
+            collect_data_config = default_config,
         )        
         self.dpgen_op_3 = ConcurrentLearning(
             self.name,
             self.block_cl_op_3,
             upload_python_package = upload_python_package,
-            image = default_image,
+            step_config = default_config,
         )
 
 
@@ -457,7 +472,7 @@ class TestLoopRestart(unittest.TestCase):
         
         self.template_script = mocked_template_script
 
-        self.type_map = []
+        self.type_map = ['H', 'O']
 
         self.incar = Path('incar')
         self.incar.write_text(mocked_incar_template)
@@ -468,6 +483,9 @@ class TestLoopRestart(unittest.TestCase):
             self.incar,
             {'foo': self.potcar},
         )
+        self.vasp_inputs_fname = Path('vasp_inputs.dat')
+        self.vasp_inputs_arti = upload_artifact(
+            dump_object_to_file(self.vasp_inputs, self.vasp_inputs_fname))
 
         self.scheduler_0 = ExplorationScheduler()        
         self.trust_level = TrustLevel(0.1, 0.3)
@@ -487,9 +505,8 @@ class TestLoopRestart(unittest.TestCase):
             max_numb_iter = 2,
         )
         self.scheduler_0.add_stage_scheduler(stage_scheduler)
-        with open('in_scheduler_0.dat', 'wb') as fp:
-            pickle.dump(self.scheduler_0, fp)
-        self.scheduler_0_artifact = upload_artifact('in_scheduler_0.dat')
+        self.scheduler_0_artifact = upload_artifact(
+            dump_object_to_file(self.scheduler_0, 'in_scheduler_0.dat'))
 
         self.scheduler_1 = ExplorationScheduler()        
         self.trust_level = TrustLevel(0.1, 0.3)
@@ -510,9 +527,8 @@ class TestLoopRestart(unittest.TestCase):
         )
         self.scheduler_1.add_stage_scheduler(stage_scheduler)        
         self.scheduler_1.add_stage_scheduler(stage_scheduler)
-        with open('in_scheduler_1.dat', 'wb') as fp:
-            pickle.dump(self.scheduler_1, fp)
-        self.scheduler_1_artifact = upload_artifact('in_scheduler_1.dat')
+        self.scheduler_1_artifact = upload_artifact(
+            dump_object_to_file(self.scheduler_1, 'in_scheduler_1.dat'))
         
         
     def setUp(self):
@@ -541,6 +557,9 @@ class TestLoopRestart(unittest.TestCase):
         for ii in [Path('scheduler.dat'), Path('scheduler_new.dat')]:
             if ii.is_file():
                 os.remove(ii)
+        for ii in [self.vasp_inputs_fname]:
+            if ii.is_file():
+                os.remove(ii)
 
     def get_restart_step(
             self,
@@ -569,11 +588,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -641,11 +660,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_1_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -666,9 +685,8 @@ class TestLoopRestart(unittest.TestCase):
         download_artifact(step.outputs.artifacts["iter_data"], path = 'iter_data')
         download_artifact(step.outputs.artifacts["models"], path = Path('models')/self.name)
         # scheduler = jsonpickle.decode(step.outputs.parameters['exploration_scheduler'].value)
-        download_artifact(step.outputs.artifacts['exploration_scheduler'])
-        with open('scheduler.dat', 'rb') as fp:
-            scheduler = pickle.load(fp)
+        scheduler = load_object_from_file(
+            download_artifact(step.outputs.artifacts["exploration_scheduler"])[0])
         self.assertEqual(scheduler.get_stage(), 2)
         self.assertEqual(scheduler.get_iteration(), 1)
         
@@ -716,11 +734,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -779,16 +797,14 @@ class TestLoopRestart(unittest.TestCase):
         step_scheduler = steps_0.pop(scheduler_idx)
         self.assertEqual(step_scheduler['phase'], 'Succeeded')
         # old_scheduler = jsonpickle.decode(step_scheduler.outputs.parameters['exploration_scheduler'].value)
-        download_artifact(step_scheduler.outputs.artifacts['exploration_scheduler'])
-        with open('scheduler.dat', 'rb') as fp:
-            old_scheduler = pickle.load(fp)
+        old_scheduler = load_object_from_file(            
+            download_artifact(step_scheduler.outputs.artifacts['exploration_scheduler'])[0])
         self.assertEqual(old_scheduler.get_stage(), 0)
         # update a stage scheduler
         old_scheduler.stage_schedulers[1] = self.scheduler_1.stage_schedulers[1]        
         # step_scheduler.modify_output_parameter("exploration_scheduler", old_scheduler)
-        with open('scheduler_new.dat', 'wb') as fp:
-            pickle.dump(old_scheduler, fp)
-        old_scheduler_artifact = upload_artifact('scheduler_new.dat')
+        old_scheduler_artifact = upload_artifact(
+            dump_object_to_file(old_scheduler, 'scheduler_new.dat'))
         step_scheduler.modify_output_artifact("exploration_scheduler", old_scheduler_artifact)
         steps_0.append(step_scheduler)
 
@@ -801,11 +817,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -826,9 +842,8 @@ class TestLoopRestart(unittest.TestCase):
         download_artifact(step.outputs.artifacts["iter_data"], path = 'iter_data')
         download_artifact(step.outputs.artifacts["models"], path = Path('models')/self.name)
         # scheduler = jsonpickle.decode(step.outputs.parameters['exploration_scheduler'].value)
-        download_artifact(step.outputs.artifacts['exploration_scheduler'])
-        with open('scheduler.dat', 'rb') as fp:
-            scheduler = pickle.load(fp)
+        scheduler = load_object_from_file(
+            download_artifact(step.outputs.artifacts["exploration_scheduler"])[0])
         self.assertEqual(scheduler.get_stage(), 2)
         self.assertEqual(scheduler.get_iteration(), 1)
         
@@ -875,11 +890,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -943,11 +958,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -968,9 +983,8 @@ class TestLoopRestart(unittest.TestCase):
         download_artifact(step.outputs.artifacts["iter_data"], path = 'iter_data')
         download_artifact(step.outputs.artifacts["models"], path = Path('models')/self.name)
         # scheduler = jsonpickle.decode(step.outputs.parameters['exploration_scheduler'].value)
-        download_artifact(step.outputs.artifacts['exploration_scheduler'])
-        with open('scheduler.dat', 'rb') as fp:
-            scheduler = pickle.load(fp)
+        scheduler = load_object_from_file(
+            download_artifact(step.outputs.artifacts["exploration_scheduler"])[0])
         self.assertEqual(scheduler.get_stage(), 2)
         self.assertEqual(scheduler.get_iteration(), 1)
         
@@ -1029,11 +1043,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -1095,11 +1109,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -1120,9 +1134,8 @@ class TestLoopRestart(unittest.TestCase):
         download_artifact(step.outputs.artifacts["iter_data"], path = 'iter_data')
         download_artifact(step.outputs.artifacts["models"], path = Path('models')/self.name)
         # scheduler = jsonpickle.decode(step.outputs.parameters['exploration_scheduler'].value)
-        download_artifact(step.outputs.artifacts['exploration_scheduler'])
-        with open('scheduler.dat', 'rb') as fp:
-            scheduler = pickle.load(fp)
+        scheduler = load_object_from_file(
+            download_artifact(step.outputs.artifacts["exploration_scheduler"])[0])
         self.assertEqual(scheduler.get_stage(), 2)
         self.assertEqual(scheduler.get_iteration(), 1)
         
@@ -1183,11 +1196,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -1248,11 +1261,11 @@ class TestLoopRestart(unittest.TestCase):
                 "template_script" : self.template_script,
                 "train_config" : {},
                 "lmp_config" : {},
-                'fp_inputs' : self.vasp_inputs,
                 "fp_config" : {},
             },
             artifacts = {
                 "exploration_scheduler" : self.scheduler_0_artifact,
+                'fp_inputs' : self.vasp_inputs_arti,
                 "init_models" : self.init_models,
                 "init_data" : self.init_data,
                 "iter_data" : self.iter_data,
@@ -1273,9 +1286,8 @@ class TestLoopRestart(unittest.TestCase):
         download_artifact(step.outputs.artifacts["iter_data"], path = 'iter_data')
         download_artifact(step.outputs.artifacts["models"], path = Path('models')/self.name)
         # scheduler = jsonpickle.decode(step.outputs.parameters['exploration_scheduler'].value)
-        download_artifact(step.outputs.artifacts['exploration_scheduler'])
-        with open('scheduler.dat', 'rb') as fp:
-            scheduler = pickle.load(fp)
+        scheduler = load_object_from_file(
+            download_artifact(step.outputs.artifacts["exploration_scheduler"])[0])
         self.assertEqual(scheduler.get_stage(), 2)
         self.assertEqual(scheduler.get_iteration(), 1)
         
